@@ -16,7 +16,7 @@ prod.spline <- function(x,
                         xeval=NULL,
                         zeval=NULL,
                         knots=c("quantiles","uniform"),
-                        basis=c("additive","tensor"),
+                        basis=c("additive","tensor","glp"),
                         deriv.index=1,
                         deriv=0) {
 
@@ -26,10 +26,10 @@ prod.spline <- function(x,
   if(missing(x) || missing (K)) stop(" must provide x and K")
   if(!is.matrix(K)) stop(" K must be a two-column matrix")
 
-  ## Additive models have intercept=FALSE in gsl.bs but intercept=TRUE
-  ## in lm()
+  ## Additive and glp models have intercept=FALSE in gsl.bs but
+  ## intercept=TRUE in lm()
 
-  gsl.intercept <- ifelse(basis=="additive", FALSE, TRUE)
+  gsl.intercept <- ifelse(basis=="additive" || basis=="glp", FALSE, TRUE)
 
   ## Care in passing (extra cast) and ensure K is a matrix of integers
   ## (K contains the spline degree [integer] for each dimension in
@@ -123,6 +123,15 @@ prod.spline <- function(x,
       dim.P.no.tensor <- NCOL(P)
       ## Solely tensor if basis==tensor      
       if(basis=="tensor") P <- tensor.prod.model.matrix(tp)
+      if(basis=="glp") {
+        P <- glp.model.matrix(tp)
+        if(deriv!=0) {
+          P.deriv <- list()
+          for(i in 1:length(tp)) P.deriv[[i]] <- matrix(0,1,ncol(tp[[i]]))
+          P.deriv[[deriv.index]] <- matrix(NA,1,ncol(tp[[deriv.index]]))
+          P[,!is.na(as.numeric(glp.model.matrix(P.deriv)))] <- 0
+        }
+      }
     } else {
       P <- tp[[1]]
       dim.P.no.tensor <- NCOL(P)
@@ -150,18 +159,18 @@ predict.kernel.spline <- function(x,
                                   z=NULL,
                                   K,
                                   lambda=NULL,
-                                  kernel.type=c("nominal","ordinal"),
+                                  is.ordered.z=NULL,
                                   xeval=NULL,
                                   zeval=NULL,
                                   knots=c("quantiles","uniform"),
-                                  basis=c("additive","tensor")){
+                                  basis=c("additive","tensor","glp")){
 
   if(missing(x) || missing(y) || missing (K)) stop(" must provide x, y and K")
   if(!is.matrix(K)) stop(" K must be a two-column matrix")  
 
   basis <- match.arg(basis)
   knots <- match.arg(knots)
-  kernel.type <- match.arg(kernel.type)
+  if(is.null(is.ordered.z)) stop(" is.ordered.z must be provided")
 
   x <- as.matrix(x)
 
@@ -179,7 +188,8 @@ predict.kernel.spline <- function(x,
       ## Degree > 0
 
       P <- prod.spline(x=x,K=K,knots=knots,basis=basis)
-      if(basis=="additive") {
+
+      if(basis=="additive" || basis=="glp") {
         model <- lm(y~P)
       } else {
         model <- lm(y~P-1)
@@ -232,10 +242,10 @@ predict.kernel.spline <- function(x,
         htt <- numeric(length=n)
         for(i in 1:nrow.z.unique) {
           zz <- ind == ind.vals[i]
-          L <- prod.kernel(Z=z,z=z.unique[ind.vals[i],],lambda=lambda,kernel.type=kernel.type)
+          L <- prod.kernel(Z=z,z=z.unique[ind.vals[i],],lambda=lambda,is.ordered.z=is.ordered.z)
           P <- prod.spline(x=x,K=K,knots=knots,basis=basis)
           k <- NCOL(P)
-          if(basis=="additive") {
+          if(basis=="additive" || basis=="glp") {
             model.z.unique <- lm(y~P,weights=L)
           } else {
             model.z.unique <- lm(y~P-1,weights=L)
@@ -263,10 +273,10 @@ predict.kernel.spline <- function(x,
         htt <- NULL ## No hatvalues for evaluation
         for(i in 1:nrow.zeval.unique) {
           zz <- ind.zeval == ind.zeval.vals[i]
-          L <- prod.kernel(Z=z,z=zeval.unique[ind.zeval.vals[i],],lambda=lambda,kernel.type=kernel.type)
+          L <- prod.kernel(Z=z,z=zeval.unique[ind.zeval.vals[i],],lambda=lambda,is.ordered.z=is.ordered.z)
           P <- prod.spline(x=x,K=K,knots=knots,basis=basis)
           k <- NCOL(P)
-          if(basis=="additive") {
+          if(basis=="additive" || basis=="glp") {
             model.z.unique <- lm(y~P,weights=L)
           } else {
             model.z.unique <- lm(y~P-1,weights=L)
@@ -296,8 +306,8 @@ predict.kernel.spline <- function(x,
         htt <- numeric(length=n)
         for(i in 1:nrow.z.unique) {
           zz <- ind == ind.vals[i]
-          L <- prod.kernel(Z=z,z=z.unique[ind.vals[i],],lambda=lambda,kernel.type=kernel.type)
-          if(basis=="additive") {
+          L <- prod.kernel(Z=z,z=z.unique[ind.vals[i],],lambda=lambda,is.ordered.z=is.ordered.z)
+          if(basis=="additive" || basis=="glp") {
             model.z.unique <- lm(y~.,weights=L,data=data.frame(y,z.factor))
           } else {
             model.z.unique <- lm(y~.-1,weights=L,data=data.frame(y,z.factor))
@@ -328,8 +338,8 @@ predict.kernel.spline <- function(x,
         htt <- NULL # no hatvalues for evaluation
         for(i in 1:nrow.zeval.unique) {
           zz <- ind.zeval == ind.zeval.vals[i]
-          L <- prod.kernel(Z=z,z=zeval.unique[ind.zeval.vals[i],],lambda=lambda,kernel.type=kernel.type)
-          if(basis=="additive") {
+          L <- prod.kernel(Z=z,z=zeval.unique[ind.zeval.vals[i],],lambda=lambda,is.ordered.z=is.ordered.z)
+          if(basis=="additive" || basis=="glp") {
             model.z.unique <- lm(y~.,weights=L,data=data.frame(y,z.factor))
           } else {
             model.z.unique <- lm(y~.-1,weights=L,data=data.frame(y,z.factor))
@@ -364,11 +374,11 @@ deriv.kernel.spline <- function(x,
                                 z=NULL,
                                 K,
                                 lambda=NULL,
-                                kernel.type=c("nominal","ordinal"),
+                                is.ordered.z=NULL,
                                 xeval=NULL,
                                 zeval=NULL,
                                 knots=c("quantiles","uniform"),
-                                basis=c("additive","tensor"),
+                                basis=c("additive","tensor","glp"),
                                 deriv.index=1,
                                 deriv=0) {
 
@@ -379,7 +389,7 @@ deriv.kernel.spline <- function(x,
 
   basis <- match.arg(basis)
   knots <- match.arg(knots)
-  kernel.type <- match.arg(kernel.type)
+  if(is.null(is.ordered.z)) stop(" is.ordered.z must be provided")
 
   x <- as.matrix(x)
 
@@ -387,7 +397,7 @@ deriv.kernel.spline <- function(x,
   ## univariate tensor spline bases. This is used only for setting
   ## appropriate columns for derivative computation.
   
-  if(basis=="additive") {
+  if(basis=="additive" || basis=="glp") {
     K.additive <- K
     K.additive[,1] <- ifelse(K[,1]>0,K[,1]-1,K[,1])
   }
@@ -414,10 +424,15 @@ deriv.kernel.spline <- function(x,
         deriv.spline <- P.deriv[,deriv.ind.vec,drop=FALSE]%*%(coef(model)[-1])[deriv.ind.vec]
         vcov.model <- vcov(model)[-1,-1,drop=FALSE]
         se.deriv <- sapply(1:NROW(P.deriv), function(i){ sqrt(P.deriv[i,deriv.ind.vec,drop=FALSE]%*%vcov.model[deriv.ind.vec,deriv.ind.vec]%*%t(P.deriv[i,deriv.ind.vec,drop=FALSE])) })
-      } else {
+      } else if(basis=="tensor") {
         model <- lm(y~P-1)
         deriv.spline <- P.deriv%*%coef(model)
         se.deriv <- sapply(1:NROW(P.deriv), function(i){ sqrt(P.deriv[i,,drop=FALSE]%*%vcov(model)%*%t(P.deriv[i,,drop=FALSE])) })
+      } else if(basis=="glp") {
+        model <- lm(y~P)
+        deriv.spline <- P.deriv%*%coef(model)[-1]
+        vcov.model <- vcov(model)[-1,-1,drop=FALSE]        
+        se.deriv <- sapply(1:NROW(P.deriv), function(i){ sqrt(P.deriv[i,,drop=FALSE]%*%vcov.model%*%t(P.deriv[i,,drop=FALSE])) })
       }
 
     } else {
@@ -450,7 +465,7 @@ deriv.kernel.spline <- function(x,
         se.deriv <- numeric(length=n)        
         for(i in 1:nrow.z.unique) {
           zz <- ind == ind.vals[i]
-          L <- prod.kernel(Z=z,z=z.unique[ind.vals[i],],lambda=lambda,kernel.type=kernel.type)
+          L <- prod.kernel(Z=z,z=z.unique[ind.vals[i],],lambda=lambda,is.ordered.z=is.ordered.z)
           P <- prod.spline(x=x,K=K,knots=knots,basis=basis)          
           P.deriv <- prod.spline(x=x,K=K,xeval=x[zz,,drop=FALSE],knots=knots,basis=basis,deriv.index=deriv.index,deriv=deriv)
           k <- NCOL(P)
@@ -466,10 +481,15 @@ deriv.kernel.spline <- function(x,
             deriv.spline[zz] <- P.deriv[,deriv.ind.vec,drop=FALSE]%*%(coef(model)[-1])[deriv.ind.vec]
             vcov.model <- vcov(model)[-1,-1,drop=FALSE]
             se.deriv[zz] <- sapply(1:NROW(P.deriv), function(i){ sqrt(P.deriv[i,deriv.ind.vec,drop=FALSE]%*%vcov.model[deriv.ind.vec,deriv.ind.vec]%*%t(P.deriv[i,deriv.ind.vec,drop=FALSE])) })            
-          } else {
+          } else if(basis=="tensor") {
             model <- lm(y~P-1)
             deriv.spline[zz] <- P.deriv%*%coef(model)
             se.deriv[zz] <- sapply(1:NROW(P.deriv), function(i){ sqrt(P.deriv[i,,drop=FALSE]%*%vcov(model)%*%t(P.deriv[i,,drop=FALSE])) })
+          } else if(basis=="glp") {
+            model <- lm(y~P)
+            deriv.spline[zz] <- P.deriv%*%coef(model)[-1]
+            vcov.model <- vcov(model)[-1,-1,drop=FALSE]
+            se.deriv[zz] <- sapply(1:NROW(P.deriv), function(i){ sqrt(P.deriv[i,,drop=FALSE]%*%vcov.model%*%t(P.deriv[i,,drop=FALSE])) })
           }
 
         }
@@ -489,7 +509,7 @@ deriv.kernel.spline <- function(x,
         se.deriv <- numeric(length=num.eval)        
         for(i in 1:nrow.zeval.unique) {
           zz <- ind.zeval == ind.zeval.vals[i]
-          L <- prod.kernel(Z=z,z=zeval.unique[ind.zeval.vals[i],],lambda=lambda,kernel.type=kernel.type)
+          L <- prod.kernel(Z=z,z=zeval.unique[ind.zeval.vals[i],],lambda=lambda,is.ordered.z=is.ordered.z)
           P <- prod.spline(x=x,K=K,knots=knots,basis=basis)
           P.deriv <- prod.spline(x=x,K=K,xeval=xeval[zz,,drop=FALSE],knots=knots,basis=basis,deriv.index=deriv.index,deriv=deriv)
           k <- NCOL(P)
@@ -505,10 +525,15 @@ deriv.kernel.spline <- function(x,
             deriv.spline[zz] <- P.deriv[,deriv.ind.vec,drop=FALSE]%*%(coef(model)[-1])[deriv.ind.vec]
             vcov.model <- vcov(model)[-1,-1,drop=FALSE]
             se.deriv[zz] <- sapply(1:NROW(P.deriv), function(i){ sqrt(P.deriv[i,deriv.ind.vec,drop=FALSE]%*%vcov.model[deriv.ind.vec,deriv.ind.vec]%*%t(P.deriv[i,deriv.ind.vec,drop=FALSE])) })
-          } else {
+          } else if(basis=="tensor") {
             model <- lm(y~P-1,weights=L)
             deriv.spline[zz] <- P.deriv%*%coef(model)
             se.deriv[zz] <- sapply(1:NROW(P.deriv), function(i){ sqrt(P.deriv[i,,drop=FALSE]%*%vcov(model)%*%t(P.deriv[i,,drop=FALSE])) })
+          } else if(basis=="glp") {
+            model <- lm(y~P,weights=L)
+            deriv.spline[zz] <- P.deriv%*%coef(model)[-1]
+            vcov.model <- vcov(model)[-1,-1,drop=FALSE]            
+            se.deriv[zz] <- sapply(1:NROW(P.deriv), function(i){ sqrt(P.deriv[i,,drop=FALSE]%*%vcov.model%*%t(P.deriv[i,,drop=FALSE])) })
           }
 
         }
@@ -539,7 +564,7 @@ deriv.kernel.spline <- function(x,
 ## This function returns the fitted/predicted values using Friedman's
 ## MARS idea of indicator function bases for categorical variables
 ## (though Friedman's MARS is much more restrictive than the setup we
-## consider here as it uses piecewise linear splines). My additional
+## consider here as it uses piece-wise linear splines). My additional
 ## twist is, as for the basis splines, that we allow a variable to not
 ## enter via a basis of zero length.
 
@@ -551,7 +576,7 @@ predict.factor.spline <- function(x,
                                   xeval=NULL,
                                   zeval=NULL,
                                   knots=c("quantiles","uniform"),
-                                  basis=c("additive","tensor"),
+                                  basis=c("additive","tensor","glp"),
                                   prune=FALSE,
                                   prune.index=NULL,
                                   trace=0){
@@ -580,14 +605,14 @@ predict.factor.spline <- function(x,
 
     if(prune && is.null(prune.index)) {
 
-      ## Pruning via stepwise CV but returning the pruned model only
+      ## Pruning via step-wise CV but returning the pruned model only
       ## if the cross-validation score is improved (lower). We create
       ## a data frame so that we can readily determine columns that
       ## have been removed and assign logical values to all columns in
       ## P.
       P.df <- data.frame(P)
       names(P.df) <- paste("P",seq(1,NCOL(P.df)),sep="")
-      if(basis=="additive") {
+      if(basis=="additive" || basis=="glp") {
         model <- lm(y~.,data=P.df)
       } else {
         model <- lm(y~.-1,data=P.df)
@@ -595,7 +620,7 @@ predict.factor.spline <- function(x,
       cv <- mean(residuals(model)^2/(1-hatvalues(model))^2)
       console <- printClear(console)
       console <- printPush("Pruning...",console = console)
-      if(basis=="additive") {
+      if(basis=="additive" || basis=="glp") {
         model.pruned <- stepCV(lm(y~.,data=P.df),
                                scope=list(upper=~.,lower=~1),
                                k=log(length(y)),
@@ -610,7 +635,7 @@ predict.factor.spline <- function(x,
       if(cv.pruned <= cv) {
         IND <- logical()
         for(i in 1:NCOL(P.df)) IND[i] <- any(names(P.df)[i]==names(model.pruned$model[,-1,drop=FALSE]))
-        if(basis=="additive") {
+        if(basis=="additive" || basis=="glp") {
           model <- lm(y~P[,IND,drop=FALSE])
         } else {
           model <- lm(y~P[,IND,drop=FALSE]-1)
@@ -618,7 +643,7 @@ predict.factor.spline <- function(x,
       } else {
         warning(" pruned model did not lower cross-validation score, using non-pruned bases")
         IND <- !logical(length=NCOL(P))
-        if(basis=="additive") {
+        if(basis=="additive" || basis=="glp") {
           model <- lm(y~P)
         } else {
           model <- lm(y~P-1)
@@ -627,7 +652,7 @@ predict.factor.spline <- function(x,
     } else if(prune) {
       ## Pruning, index passed in...
       IND <- prune.index
-      if(basis=="additive") {
+      if(basis=="additive" || basis=="glp") {
         model <- lm(y~P[,IND,drop=FALSE])
       } else {
         model <- lm(y~P[,IND,drop=FALSE]-1)
@@ -637,7 +662,7 @@ predict.factor.spline <- function(x,
     } else {
       ## No pruning
       IND <- !logical(length=NCOL(P))
-      if(basis=="additive") {
+      if(basis=="additive" || basis=="glp") {
         model <- lm(y~P)
       } else {
         model <- lm(y~P-1)
@@ -689,7 +714,7 @@ predict.factor.spline <- function(x,
 ## This function returns the fitted/predicted values using Friedman's
 ## MARS idea of indicator function bases for categorical variables
 ## (though Friedman's MARS is much more restrictive than the setup we
-## consider here as it uses piecewise linear splines). My additional
+## consider here as it uses piece-wise linear splines). My additional
 ## twist is, as for the basis splines, that we allow a variable to not
 ## enter via a basis of zero length.
 
@@ -701,7 +726,7 @@ deriv.factor.spline <- function(x,
                                 xeval=NULL,
                                 zeval=NULL,
                                 knots=c("quantiles","uniform"),
-                                basis=c("additive","tensor"),
+                                basis=c("additive","tensor","glp"),
                                 deriv.index=1,
                                 deriv=0,
                                 prune.index=NULL) {
@@ -719,7 +744,7 @@ deriv.factor.spline <- function(x,
   ## univariate tensor spline bases. This is used only for setting
   ## appropriate columns for derivative computation.
   
-  if(basis=="additive") {
+  if(basis=="additive" || basis=="glp") {
     K.additive <- K
     K.additive[,1] <- ifelse(K[,1]>0,K[,1]-1,K[,1])
   }
@@ -753,12 +778,18 @@ deriv.factor.spline <- function(x,
       deriv.end <- deriv.start+sum(K.additive[deriv.index,])-1
       deriv.ind.vec[deriv.start:deriv.end] <- TRUE
       deriv.ind.vec <- ifelse(prune.index,deriv.ind.vec,FALSE)
-    } else {
+    } else if(basis=="tensor") {
       model <- lm(y~P[,prune.index,drop=FALSE]-1)
       coef.vec.model[prune.index] <- coef(model)
       vcov.mat.model[prune.index,prune.index] <- vcov(model)
       deriv.ind.vec[1:dim.P.tensor] <- TRUE            
       deriv.ind.vec <- ifelse(prune.index,deriv.ind.vec,FALSE)
+    } else if(basis=="glp") {
+      model <- lm(y~P[,prune.index,drop=FALSE])
+      coef.vec.model[prune.index] <- coef(model)[-1]
+      vcov.mat.model[prune.index,prune.index] <- vcov(model)[-1,-1,drop=FALSE]
+      deriv.ind.vec[1:dim.P.tensor] <- TRUE            
+      deriv.ind.vec <- ifelse(prune.index,deriv.ind.vec,FALSE)
     }
 
     deriv.spline <- P.deriv[,deriv.ind.vec,drop=FALSE]%*%coef.vec.model[deriv.ind.vec]
@@ -784,6 +815,13 @@ deriv.factor.spline <- function(x,
 ## calculate the leave-one-out cross-validation function for
 ## categorical kernel splines.
 
+## June 24 2011 - replaced lm() with model <- lm.fit/lm.wfit and
+## hat(model$qr) both here and in cv.factor.spline. Here it reduces
+## runtime by 20-30%. But more importantly lm.fit is more `robust'
+## than lsfit (lm.fit is the `workhorse' of lm, lsfit calls LAPACK
+## code). Note that it is noticeable as it returns a larger cv value
+## for more complicated problems which is naturally desirable.
+
 cv.kernel.spline <- function(x,
                              y,
                              z=NULL,
@@ -793,9 +831,9 @@ cv.kernel.spline <- function(x,
                              ind,
                              ind.vals,
                              nrow.z.unique,
-                             kernel.type=c("nominal","ordinal"),
+                             is.ordered.z=NULL,
                              knots=c("quantiles","uniform"),
-                             basis=c("additive","tensor"),
+                             basis=c("additive","tensor","glp"),
                              cv.func=c("cv.ls","cv.gcv","cv.aic")) {
 
   if(missing(x) || missing(y) || missing (K)) stop(" must provide x, y and K")
@@ -803,38 +841,28 @@ cv.kernel.spline <- function(x,
   if(!is.matrix(K)) stop(" K must be a two-column matrix")  
 
   basis <- match.arg(basis)
-  kernel.type <- match.arg(kernel.type)
+  if(is.null(is.ordered.z)) stop(" is.ordered.z must be provided")
   knots <- match.arg(knots)
   cv.func <- match.arg(cv.func)
 
   ## Additive spline regression models have an intercept in the lm()
   ## model (though not in the gsl.bs function)
 
-  lm.intercept <- ifelse(basis=="additive", TRUE, FALSE)
+  lm.intercept <- ifelse(basis=="additive" || basis=="glp", TRUE, FALSE)
 
   ## Without computing P, compute the number of columns that P would
   ## be and if degrees of freedom is 1 or less, return a large penalty.
 
   n <- length(y)
-  c.vec <- NULL
-
-  if(basis=="additive") {
-    if(any(K[,1] > 0)) c.vec <- rowSums(K[K[,1]!=0,,drop=FALSE])-1
-    ncol.P <- sum(c.vec)
-  }
-  if(basis=="tensor") {
-    if(any(K[,1] > 0)) c.vec <- rowSums(K[K[,1]!=0,,drop=FALSE])
-    ncol.P <- prod(c.vec)
-  }
-
-  if(ncol.P >= (n-1)) return(.Machine$double.xmax)
 
   ## Otherwise, compute the cross-validation function
 
   if(is.null(z)) {
     ## No categorical predictors
     if(any(K[,1] > 0)) {
-      suppressWarnings(epsilon <- residuals(lsfit(P <- prod.spline(x=x,K=K,knots=knots,basis=basis),y,intercept=lm.intercept)))
+      P <- prod.spline(x=x,K=K,knots=knots,basis=basis)
+      if(NCOL(P) >= (n-1)) return(sqrt(.Machine$double.xmax))
+      suppressWarnings(epsilon <- residuals(lsfit(P,y,intercept=lm.intercept)))
       htt <- hat(P)
       htt <- ifelse(htt == 1, 1-.Machine$double.eps, htt)      
     } else {
@@ -852,31 +880,39 @@ cv.kernel.spline <- function(x,
       cv <- ifelse(penalty < 0, .Machine$double.xmax, log(sigmasq)+penalty);
     }
   } else {
-    ## Categorical predictors
+    ## Categorical predictors - this is the workhorse
     z <- as.matrix(z)
     num.z <- NCOL(z)
     epsilon <- numeric(length=n)
     htt <- numeric(length=n)
     if(any(K[,1] > 0)) {
-      P <- prod.spline(x=x,K=K,knots=knots,basis=basis) ## moved outside loop Feb 3 2011
+      P <- prod.spline(x=x,K=K,knots=knots,basis=basis)
+      ## Check for 1 or fewer degrees of freedom
+      if(NCOL(P) >= (n-1)) return(sqrt(.Machine$double.xmax))
       for(i in 1:nrow.z.unique) {
         zz <- ind == ind.vals[i]
-        L <- prod.kernel(Z=z,z=z.unique[ind.vals[i],],lambda=lambda,kernel.type=kernel.type)
-        if(basis=="additive") {
-          model <- lm(y~P,weights=L)
+        L <- prod.kernel(Z=z,z=z.unique[ind.vals[i],],lambda=lambda,is.ordered.z=is.ordered.z)
+        if(basis=="additive" || basis=="glp") {
+          ## Additive spline regression models have an intercept in the lm()
+          ## model (though not in the gsl.bs function)
+          model <- lm.wfit(cbind(1,P),y,L)
+          ## Check for rank-deficient fit
+          if(model$rank != (NCOL(P)+1)) return(sqrt(.Machine$double.xmax))
         } else {
-          model <- lm(y~P-1,weights=L)
+          model <- lm.wfit(P,y,L)
+          ## Check for rank-deficient fit          
+          if(model$rank != NCOL(P)) return(sqrt(.Machine$double.xmax))
         }
         epsilon[zz] <- residuals(model)[zz]
-        htt[zz] <- hatvalues(model)[zz]
+        htt[zz] <- hat(model$qr)[zz]
       }
     } else {
       z.factor <- data.frame(factor(z[,1]))
       if(num.z > 1) for(i in 2:num.z) z.factor <- data.frame(z.factor,factor(z[,i]))
       for(i in 1:nrow.z.unique) {
         zz <- ind == ind.vals[i]
-        L <- prod.kernel(Z=z,z=z.unique[ind.vals[i],],lambda=lambda,kernel.type=kernel.type)
-        if(basis=="additive") {
+        L <- prod.kernel(Z=z,z=z.unique[ind.vals[i],],lambda=lambda,is.ordered.z=is.ordered.z)
+        if(basis=="additive" || basis=="glp") {
           model <- lm(y~.,weights=L,data=data.frame(y,z.factor))
         } else {
           model <- lm(y~.-1,weights=L,data=data.frame(y,z.factor))
@@ -889,9 +925,9 @@ cv.kernel.spline <- function(x,
     htt <- ifelse(htt == 1, 1-.Machine$double.eps, htt)
 
     if(cv.func == "cv.ls") {
-      cv <- mean(epsilon^2/(1-htt)^2)
+      cv <- mean((epsilon/(1-htt))^2)
     } else if(cv.func == "cv.gcv"){
-      cv <- mean(epsilon^2/(1-mean(htt))^2)
+      cv <- mean((epsilon/(1-mean(htt)))^2)
     } else if(cv.func == "cv.aic"){
       sigmasq <- mean(epsilon^2)
       traceH <- sum(htt)
@@ -915,70 +951,49 @@ cv.factor.spline <- function(x,
                              z=NULL,
                              K,
                              I=NULL,
-                             kernel.type=c("nominal","ordinal"),
                              knots=c("quantiles","uniform"),
-                             basis=c("additive","tensor"),
+                             basis=c("additive","tensor","glp"),
                              cv.func=c("cv.ls","cv.gcv","cv.aic")) {
 
   if(missing(x) || missing(y) || missing (K)) stop(" must provide x, y and K")
   if(!is.matrix(K)) stop(" K must be a two-column matrix")  
 
   basis <- match.arg(basis)
-  kernel.type <- match.arg(kernel.type)
   knots <- match.arg(knots)
 
   cv.func <- match.arg(cv.func)
 
   n <- NROW(x)
 
-  ## Additive spline regression models have an intercept in the lm()
-  ## model (though not in the gsl.bs function)
-
-  lm.intercept <- ifelse(basis=="additive", TRUE, FALSE)
-
-  ## Without computing P, compute the number of columns that P would
-  ## be and if degrees of freedom is 1 or less, return a large penalty.
-
-  c.vec <- NULL
-  i.vec <- NULL
-
-  if(any(I > 0)) {
-    i.vec <- numeric()
-    j <- 1
-    for(i in 1:ncol(z)) {
-      if(I[i] != 0) {
-        i.vec[j] <- length(unique(z[,i]))-1
-        j <- j + 1
-      }
-    }
-  }
-
-  if(basis=="additive") {
-    if(any(K[,1] > 0)) c.vec <- rowSums(K[K[,1]!=0,,drop=FALSE])-1
-    ncol.P <- sum(c(c.vec,i.vec))
-  }
-  if(basis=="tensor") {
-    if(any(K[,1] > 0)) c.vec <- rowSums(K[K[,1]!=0,,drop=FALSE])
-    ncol.P <- prod(c(c.vec,i.vec))
-  }
-
-  if(ncol.P >= (n-1)) return(.Machine$double.xmax)
-
   ## Otherwise, compute the cross-validation function
 
   if(any(K[,1] > 0)||any(I > 0)) {
-    suppressWarnings(epsilon <- residuals(lsfit(P <- prod.spline(x=x,z=z,K=K,I=I,knots=knots,basis=basis),y,intercept=lm.intercept)))
-    htt <- hat(P)
+    P <- prod.spline(x=x,z=z,K=K,I=I,knots=knots,basis=basis)
+    ## Check for 1 or fewer degrees of freedom
+    if(NCOL(P) >= (n-1)) return(sqrt(.Machine$double.xmax))
+    if(basis=="additive" || basis=="glp") {
+      ## Additive spline regression models have an intercept in the
+      ## lm() model (though not in the gsl.bs function)
+      model <- lm.fit(cbind(1,P),y)
+      ## Check for rank-deficient fit
+      if(model$rank != (NCOL(P)+1)) return(sqrt(.Machine$double.xmax))
+    } else {
+      model <- lm.fit(P,y)
+      ## Check for rank-deficient fit      
+      if(model$rank != NCOL(P)) return(sqrt(.Machine$double.xmax))      
+    }
+    htt <- hat(model$qr)
     htt <- ifelse(htt == 1, 1-.Machine$double.eps, htt)
+    epsilon <- residuals(model)
   } else {
     htt <- rep(1/n,n)
     epsilon <- y-mean(y)
   }
 
   if(cv.func == "cv.ls") {
-    cv <- mean(epsilon^2/(1-htt)^2)
+    cv <- mean((epsilon/(1-htt))^2)
   } else if(cv.func == "cv.gcv"){
-    cv <- mean(epsilon^2/(1-mean(htt))^2)
+    cv <- mean((epsilon/(1-mean(htt)))^2)
   } else if(cv.func == "cv.aic"){
     sigmasq <- mean(epsilon^2)
     traceH <- sum(htt)
